import os
import numpy as np
import pytest

from rail.utils.path_utils import find_rail_file
from rail.interfaces import ToolFactory


def test_tool_factory():

    stage = ToolFactory.build_tool_stage(
        'pq_to_hdf',
        'TableConverter',
        'rail.tools.table_tools',
        'dummy.in',
        output_format='numpyDict',
    )

    input_file = find_rail_file('examples_data/testdata/rubin_dm_dc2_example.pq')
    
    out_handle = ToolFactory.run_tool_stage(
        stage,
        input_file,
    )

    check_stage = ToolFactory.get_tool_stage('pq_to_hdf')
    assert check_stage == stage

    with pytest.raises(KeyError):
        ToolFactory.get_tool_stage('nope')    
    
    ToolFactory.reset()
    assert not ToolFactory._stage_dict
    
    try:
        os.unlink('inprogress_output_pq_to_hdf.hdf5')
    except:
        pass
    try:
        os.unlink('output_pq_to_hdf.hdf5')
    except:
        pass

    
    
    
    
        



"""Core code for RAIL"""

from .stage import RailPipeline, RailStage
from .introspection import RailEnv

__all__ = [
    "RailPipeline",
    "RailStage",
    "RailEnv",
]


def find_version():
    """Find the version"""
    # setuptools_scm should install a
    # file _version alongside this one.
    from . import _version  # pylint: disable=import-outside-toplevel

    return _version.version


try:
    __version__ = find_version()
except ImportError:  # pragma: no cover
    __version__ = "unknown"



"""
A summarizer that simple makes a histogram of a point estimate
"""

import numpy as np
import qp

from ceci.config import StageParameter as Param
from rail.estimation.summarizer import PZSummarizer
from rail.estimation.informer import PzInformer
from rail.core.data import QPHandle, TableHandle


class NaiveStackInformer(PzInformer):
    """Placeholder Informer"""

    name = "NaiveStackInformer"
    config_options = PzInformer.config_options.copy()

    def run(self):
        self.add_data("model", np.array([None]))


class NaiveStackSummarizer(PZSummarizer):
    """Summarizer which stacks individual P(z)"""

    name = "NaiveStackSummarizer"
    config_options = PZSummarizer.config_options.copy()
    config_options.update(
        zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        nsamples=Param(int, 1000, msg="Number of sample distributions to create"),
    )
    inputs = [("input", QPHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]

    def __init__(self, args, **kwargs):
        super().__init__(args, **kwargs)
        self.zgrid = None

    def _setup_iterator(self):
        itr = self.input_iterator("input")
        for s, e, d in itr:
            yield s, e, d, np.ones(e-s, dtype=bool)


    def run(self):
        handle = self.get_handle("input", allow_missing=True)
        self._input_length = handle.size()
        iterator = self._setup_iterator()
        self.zgrid = np.linspace(
            self.config.zmin, self.config.zmax, self.config.nzbins + 1
        )
        # Initiallizing the stacking pdf's
        yvals = np.zeros((1, len(self.zgrid)))
        bvals = np.zeros((self.config.nsamples, len(self.zgrid)))
        bootstrap_matrix = self._broadcast_bootstrap_matrix()

        first = True
        for s, e, test_data, mask in iterator:
            print(f"Process {self.rank} running estimator on chunk {s} - {e}")
            self._process_chunk(s, e, test_data, mask, first, bootstrap_matrix, yvals, bvals)
            first = False
        if self.comm is not None:  # pragma: no cover
            bvals, yvals = self._join_histograms(bvals, yvals)

        if self.rank == 0:
            sample_ens = qp.Ensemble(
                qp.interp, data=dict(xvals=self.zgrid, yvals=bvals)
            )
            qp_d = qp.Ensemble(qp.interp, data=dict(xvals=self.zgrid, yvals=yvals))
            self.add_data("output", sample_ens)
            self.add_data("single_NZ", qp_d)

    def _process_chunk(self, start, end, data, mask, _first, bootstrap_matrix, yvals, bvals):
        pdf_vals = data.pdf(self.zgrid)
        squeeze_mask = np.squeeze(mask)
        yvals += np.expand_dims(
            np.sum(np.where(np.isfinite(pdf_vals[squeeze_mask,:]), pdf_vals[squeeze_mask], 0.0), axis=0), 0
        )
        # qp_d is the normalized probability of the stack, we need to know how many galaxies were
        for i in range(self.config.nsamples):
            bootstrap_draws = bootstrap_matrix[:, i]
            # Neither all of the bootstrap_draws are in this chunk nor the index starts at "start"
            chunk_mask = (bootstrap_draws >= start) & (bootstrap_draws < end)
            bootstrap_draws = bootstrap_draws[chunk_mask] - start
            zarr = np.where(squeeze_mask, pdf_vals.T, 0.).T[bootstrap_draws]
            bvals[i] += np.sum(zarr, axis=0)


class NaiveStackMaskedSummarizer(NaiveStackSummarizer):
    name = "NaiveStackMaskedSummarizer"
    config_options = NaiveStackSummarizer.config_options.copy()
    config_options.update(
        selected_bin=Param(int, -1, msg="bin to use"),
    )
    inputs = [("input", QPHandle), ("tomography_bins", TableHandle)]
    outputs = [("output", QPHandle), ("single_NZ", QPHandle)]


    def _setup_iterator(self):

        selected_bin = self.config.selected_bin
        if self.config.tomography_bins in ['none', None]:
            selected_bin = -1

        if selected_bin == -1:
            itrs = [self.input_iterator('input')]
        else:
            itrs = [self.input_iterator('input'), self.input_iterator('tomography_bins')]

        for it in zip(*itrs):
            first = True
            mask = None
            for s, e, d in it:
                if first:
                    start = s
                    end = e
                    pz_data = d
                    first = False
                else:
                    mask = d['class_id'] == self.config.selected_bin
            if mask is None:
                mask = np.ones(pz_data.npdf, dtype=bool)
            yield start, end, pz_data, mask

    def summarize(self, input_data, tomo_bins=None):
        """Override the Summarizer.summarize() method to take tomo bins
        as an additional input

        Parameters
        ----------
        input_data : `qp.Ensemble`
            Per-galaxy p(z), and any ancilary data associated with it

        tomo_bins : `table-like`
            Tomographic bins file

        Returns
        -------
        output: `qp.Ensemble`
            Ensemble with n(z), and any ancilary data
        """
        self.set_data("input", input_data)
        if tomo_bins is None:
            self.config.tomography_bins = None
        else:
            self.set_data("tomography_bins", tomo_bins)
        self.run()
        self.finalize()
        return self.get_handle("output")

            



"""Rail-specific data management"""

import os
import pickle
import enum
import tables_io
import qp

from .model import Model


class DataHandle:  # pylint: disable=too-many-instance-attributes
    """Class to act as a handle for a bit of data.  Associating it with a file and
    providing tools to read & write it to that file

    Parameters
    ----------
    tag : str
        The tag under which this data handle can be found in the store
    data : any or None
        The associated data
    path : str or None
        The path to the associated file
    creator : str or None
        The name of the stage that created this data handle
    """

    suffix = ""

    # This is to keep track of all the sub-types
    data_handle_type_dict = {}

    def __init__(self, tag, data=None, path=None, creator=None):
        """Constructor"""
        self.tag = tag
        if data is not None:
            self._validate_data(data)
        self.data = data
        self.path = path
        self.creator = creator
        self.fileObj = None
        self.groups = None
        self.partial = False
        self.length = None

    def __init_subclass__(cls, **kwargs):
        """Register the subclass with the dict"""
        cls.data_handle_type_dict[cls.__name__] = cls

    @classmethod
    def get_sub_class(cls, class_name):
        """Get a particular subclass by name"""
        return cls.data_handle_type_dict[class_name]

    @classmethod
    def print_sub_classes(cls):
        """Print the list of all the subclasses"""
        for key, val in cls.data_handle_type_dict.items():
            print(f"{key}: {val}")

    def open(self, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        if self.path is None:
            raise ValueError("DataHandle.open() called but path has not been specified")
        self.fileObj = self._open(os.path.expandvars(self.path), **kwargs)
        return self.fileObj

    @classmethod
    def _open(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._open")  # pragma: no cover

    def close(self, **kwargs):  # pylint: disable=unused-argument
        """Close"""
        self.fileObj = None

    def read(self, force=False, **kwargs):
        """Read and return the data from the associated file"""
        if self.data is not None and not force:
            return self.data
        self.set_data(self._read(os.path.expandvars(self.path), **kwargs))
        return self.data

    def __call__(self, **kwargs):
        """Return the data, re-reading the fill if needed"""
        if self.has_data and not self.partial:
            return self.data
        return self.read(force=True, **kwargs)

    @classmethod
    def _read(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._read")  # pragma: no cover

    def write(self, **kwargs):
        """Write the data to the associated file"""
        if self.path is None:
            raise ValueError(
                "TableHandle.write() called but path has not been specified"
            )
        if self.data is None:
            raise ValueError(
                f"TableHandle.write() called for path {self.path} with no data"
            )
        outdir = os.path.dirname(os.path.abspath(os.path.expandvars(self.path)))
        if not os.path.exists(outdir):  # pragma: no cover
            os.makedirs(outdir, exist_ok=True)
        return self._write(self.data, os.path.expandvars(self.path), **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        raise NotImplementedError("DataHandle._write")  # pragma: no cover

    def initialize_write(self, data_length, **kwargs):
        """Initialize file to be written by chunks"""
        if self.path is None:  # pragma: no cover
            raise ValueError(
                "TableHandle.write() called but path has not been specified"
            )
        self.groups, self.fileObj = self._initialize_write(
            self.data, os.path.expandvars(self.path), data_length, **kwargs
        )

    @classmethod
    def _initialize_write(cls, data, path, data_length, **kwargs):
        raise NotImplementedError("DataHandle._initialize_write")  # pragma: no cover

    def write_chunk(self, start, end, **kwargs):
        """Write the data to the associated file"""
        if self.data is None:
            raise ValueError(
                f"TableHandle.write_chunk() called for path {self.path} with no data"
            )
        if self.fileObj is None:
            raise ValueError(
                f"TableHandle.write_chunk() called before open for {self.tag} : {self.path}"
            )
        return self._write_chunk(
            self.data, self.fileObj, self.groups, start, end, **kwargs
        )

    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        raise NotImplementedError("DataHandle._write_chunk")  # pragma: no cover

    def finalize_write(self, **kwargs):
        """Finalize and close file written by chunks"""
        if self.fileObj is None:  # pragma: no cover
            raise ValueError(
                f"TableHandle.finalize_wite() called before open for {self.tag} : {self.path}"
            )
        self._finalize_write(self.data, self.fileObj, **kwargs)

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        raise NotImplementedError("DataHandle._finalize_write")  # pragma: no cover

    def iterator(self, **kwargs):
        """Iterator over the data"""
        if self.data is not None and self.partial is False:
            return self._in_memory_iterator(**kwargs)
        return self._iterator(self.path, **kwargs)

    def set_data(self, data, partial=False):
        """Set the data for a chunk, and set the partial flag to true"""
        self._validate_data(data)
        self.data = data
        self.partial = partial

    @classmethod
    def _validate_data(cls, data):  # pylint: disable=unused-argument
        """Make sure that the right type of data is being passed in"""
        return

    def size(self, **kwargs):
        """Return the size of the data associated to this handle"""
        if self.partial:
            if self.length is not None:  # pragma: no cover
                return self.length
            return self._size(self.path, **kwargs)
        if self.data is not None:
            return self.data_size(**kwargs)
        return self._size(self.path, **kwargs)

    def _size(self, path, **kwargs):
        raise NotImplementedError("DataHandle._size")  # pragma: no cover

    def data_size(self, **kwargs):
        """Return the size of the in memorry data"""
        if self.data is None:  # pragma: no cover
            return 0
        return self._data_size(self.data, **kwargs)

    def _data_size(self, data, **kwargs):
        raise NotImplementedError("DataHandle._data_size")  # pragma: no cover

    def _in_memory_iterator(self, **kwargs):
        raise NotImplementedError("DataHandle._in_memory_iterator")  # pragma: no cover

    @classmethod
    def _iterator(cls, path, **kwargs):
        raise NotImplementedError("DataHandle._iterator")  # pragma: no cover

    @property
    def has_data(self):
        """Return true if the data for this handle are loaded"""
        return self.data is not None

    @property
    def has_path(self):
        """Return true if the path for the associated file is defined"""
        return self.path is not None

    @property
    def is_written(self):
        """Return true if the associated file has been written"""
        if self.path is None:
            return False
        return os.path.exists(os.path.expandvars(self.path))

    def __str__(self):
        s = f"{type(self)} "
        if self.has_path:
            s += f"{self.path}, ("
        else:
            s += "None, ("
        if self.is_written:
            s += "w"
        if self.has_data:
            s += "d"
        s += ")"
        return s

    @classmethod
    def make_name(cls, tag):
        """Construct and return file name for a particular data tag"""
        if cls.suffix:
            return f"{tag}.{cls.suffix}"
        return tag  # pragma: no cover


class TableHandle(DataHandle):
    """DataHandle for single tables of data"""

    suffix = None

    def set_data(self, data, partial=False):
        """Set the data for a chunk, and set the partial flag to true"""
        self._validate_data(data)
        self.data = data
        self.partial = partial

    @classmethod
    def _validate_data(cls, data):  # pylint: disable=unused-argument
        """Make sure that the right type of data is being passed in"""
        return

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file"""
        return tables_io.read(path, **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associated file"""
        return tables_io.write(data, path, **kwargs)

    def _size(self, path, **kwargs):
        if path in [None, "none", "None"]:  # pragma: no cover
            return 0
        return tables_io.io.getInputDataLength(path, **kwargs)

    def _data_size(self, data, **kwargs):
        group_name = kwargs.get("groupname", None)
        if group_name:
            try:
                data = data[group_name]
            except KeyError:  # pragma: no cover
                pass
        max_l = 0
        for _k, v in data.items():
            max_l = max(max_l, len(v))
        return max_l

    def _in_memory_iterator(self, **kwargs):
        nrows = self.data_size(**kwargs)
        groupname = kwargs.get("groupname", None)
        if isinstance(self.data, dict) and groupname:
            try:
                table = self.data[groupname]
            except KeyError:  # pragma: no cover
                table = self.data
        else:
            table = self.data
        for start, end in tables_io.ioUtils.data_ranges_by_rank(
            nrows, kwargs.get("chunk_size", 100000), 1, 0
        ):
            if isinstance(self.data, dict):
                yield start, end, tables_io.arrayUtils.sliceDict(
                    table, slice(start, end)
                )
            else:  # pragma: no cover
                yield start, end, table[start:end]

    @classmethod
    def _iterator(cls, path, **kwargs):
        """Iterate over the data"""
        return tables_io.iteratorNative(path, **kwargs)


class Hdf5Handle(TableHandle):  # pragma: no cover
    """DataHandle for a table written to HDF5"""

    suffix = "hdf5"

    @classmethod
    def _initialize_write(cls, data, path, data_length, **kwargs):
        initial_dict = cls._get_allocation_kwds(data, data_length)
        comm = kwargs.get("communicator", None)
        group, fout = tables_io.io.initializeHdf5WriteSingle(
            path, groupname=None, comm=comm, **initial_dict
        )
        return group, fout

    @classmethod
    def _get_allocation_kwds(cls, data, data_length):
        keywords = {}
        for key, array in data.items():
            shape = list(array.shape)
            if not shape:
                continue
            shape[0] = data_length
            keywords[key] = (shape, array.dtype)
        return keywords

    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        tables_io.io.writeDictToHdf5ChunkSingle(fileObj, data, start, end, **kwargs)

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        return tables_io.io.finalizeHdf5Write(fileObj, **kwargs)


class FitsHandle(TableHandle):
    """DataHandle for a table written to fits"""

    suffix = "fits"


class PqHandle(TableHandle):
    """DataHandle for a parquet table"""

    suffix = "pq"

    def _size(self, path, **kwargs):
        return tables_io.io.getInputDataLengthPq(path, **kwargs)


class QPHandle(DataHandle):
    """DataHandle for qp ensembles"""

    suffix = "hdf5"

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file"""
        return qp.read(path)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associated file"""
        return data.write_to(path)

    @classmethod
    def _initialize_write(cls, data, path, data_length, **kwargs):
        comm = kwargs.get("communicator", None)
        return data.initializeHdf5Write(path, data_length, comm)

    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        return data.writeHdf5Chunk(fileObj, start, end)

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        return data.finalizeHdf5Write(fileObj)

    @classmethod
    def _validate_data(cls, data):
        if not isinstance(data, qp.Ensemble):
            raise TypeError(
                f"Expected `data` to be a `qp.Ensemble`, but {type(data)} was provided."
                "Perhaps you meant to use `TableHandle`?"
            )

    def _size(self, path, **kwargs):
        if self.data is not None and not self.partial:  # pragma: no cover
            return self._data_size(self.data)
        if path in [None, "none", "None"]:  # pragma: no cover
            return 0
        return tables_io.io.getInputDataLengthHdf5(path, groupname="data")

    def _data_size(self, data, **kwargs):
        return self.data.npdf

    def _in_memory_iterator(self, **kwargs):
        nrows = self.data.npdf
        for start, end in tables_io.ioUtils.data_ranges_by_rank(
            nrows, kwargs.get("chunk_size", 100000), 1, 0
        ):
            yield start, end, self.data[start:end]

    @classmethod
    def _iterator(cls, path, **kwargs):
        """Iterate over the data"""
        kwargs.pop("groupname", "None")
        return qp.iterator(path, **kwargs)


class QPDictHandle(DataHandle):
    """DataHandle for dictionaries of qp ensembles"""

    suffix = "hdf5"

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the dictionary of qp.Ensembles from the associated file"""
        return qp.read_dict(path)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data (a dictionary of qp.Ensembles) to the associated file"""
        return qp.write_dict(path, data)


class QPOrTableHandle(QPHandle, Hdf5Handle):
    """DataHandle that should work with either qp.ensembles or tables"""

    suffix = "hdf5"

    class PdfOrValue(enum.Enum):
        unknown = -1
        distribution = 0
        point_estimate = 1
        both = 2

        def has_dist(self):
            return self.value in [0, 2]

        def has_point(self):
            return self.value in [1, 2]

    def is_qp(self):
        """Check if the associated data or file is a QP ensemble"""
        if self.path in [None, "None", "none"]:
            return isinstance(self.data, qp.Ensemble)
        return qp.is_qp_file(self.path)

    def check_pdf_or_point(self):
        """Check the associated file to see if it is a QP pdf, point estimate or both"""
        if self.is_qp():
            return self.PdfOrValue.both
        return self.PdfOrValue.point_estimate

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file

        Notes
        -----
        This will simply open the file and return a file-like object to the caller.
        It will not read or cache the data
        """
        return tables_io.io.io_open(path, **kwargs)  # pylint: disable=no-member

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file"""
        if qp.is_qp_file(path):
            return qp.read(path, **kwargs)
        return tables_io.read(path, **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associated file"""
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _initialize_write(cls, data, path, data_length, **kwargs):
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _write_chunk(cls, data, fileObj, groups, start, end, **kwargs):
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _finalize_write(cls, data, fileObj, **kwargs):
        raise RuntimeError(
            "QPOrTableHandle should be used for input, not output"
        )  # pragma: no cover

    @classmethod
    def _validate_data(cls, data):
        pass

    def _size(self, path, **kwargs):
        if self.is_qp():
            return QPHandle._size(self, path, **kwargs)
        return Hdf5Handle._size(self, path, **kwargs)

    def _data_size(self, data, **kwargs):
        if self.is_qp():
            return self.data.npdf
        return Hdf5Handle._data_size(self, data, **kwargs)

    def _in_memory_iterator(self, **kwargs):
        if self.is_qp():
            return QPHandle._in_memory_iterator(self, **kwargs)
        return Hdf5Handle._in_memory_iterator(self, **kwargs)

    @classmethod
    def _iterator(cls, path, **kwargs):
        """Iterate over the data"""
        if qp.is_qp_file(path):
            return QPHandle._iterator(path, **kwargs)
        return Hdf5Handle._iterator(path, **kwargs)


def default_model_read(modelfile):
    """Default function to read model files, simply used pickle.load"""
    with open(modelfile, 'rb') as fin:
        read_data = pickle.load(fin)
    if isinstance(read_data, Model):
        ret_data = read_data.data
    else:
        ret_data = read_data
    return ret_data


def default_model_write(model, path):
    """Write the model, this default implementation uses pickle"""
    with open(path, "wb") as fout:
        pickle.dump(obj=model, file=fout, protocol=pickle.HIGHEST_PROTOCOL)


class ModelDict(dict):
    """
    A specialized dict to keep track of individual estimation models objects:
    this is just a dict these additional features

    1. Keys are paths
    2. There is a read(path, force=False) method that reads a model object and
    inserts it into the dictionary
    3. There is a single static instance of this class
    """

    def open(self, path, mode, **kwargs):
        """Open the file and return the file handle"""
        return open(path, mode, **kwargs)  # pylint: disable=unspecified-encoding

    def read(
        self, path, force=False, reader=None, **kwargs
    ):  # pylint: disable=unused-argument
        """Read a model into this dict"""
        if reader is None:
            reader = default_model_read
        if force or path not in self:
            model = reader(path)
            self[path] = model
            return model
        return self[path]

    def write(
        self, model, path, force=False, writer=None, **kwargs
    ):  # pylint: disable=unused-argument
        """Write the model, this default implementation uses pickle"""
        if writer is None:
            writer = default_model_write
        if force or path not in self or not os.path.exists(path):
            self[path] = model
            writer(model, path)


class ModelHandle(DataHandle):
    """DataHandle for machine learning models"""

    suffix = "pkl"

    model_factory = ModelDict()

    @classmethod
    def _open(cls, path, **kwargs):
        """Open and return the associated file"""
        kwcopy = kwargs.copy()
        if kwcopy.pop("mode", "r") == "w":
            return cls.model_factory.open(path, mode="wb", **kwcopy)
        return cls.model_factory.read(path, **kwargs)

    @classmethod
    def _read(cls, path, **kwargs):
        """Read and return the data from the associated file"""
        return cls.model_factory.read(path, **kwargs)

    @classmethod
    def _write(cls, data, path, **kwargs):
        """Write the data to the associated file"""
        return cls.model_factory.write(data, path, **kwargs)


class DataStore(dict):
    """Class to provide a transient data store

    This class:
    1) associates data products with keys
    2) provides functions to read and write the various data produces to associated files
    """

    allow_overwrite = False

    def __init__(self, **kwargs):
        """Build from keywords

        Note
        ----
        All of the values must be data handles of this will raise a TypeError
        """
        dict.__init__(self)
        for key, val in kwargs.items():
            self[key] = val

    def __str__(self):
        """Override __str__ casting to deal with `TableHandle` objects in the map"""
        s = "{"
        for key, val in self.items():
            s += f"  {key}:{val}\n"
        s += "}"
        return s

    def __repr__(self):
        """A custom representation"""
        s = "DataStore\n"
        s += self.__str__()
        return s

    def __setitem__(self, key, value):
        """Override the __setitem__ to work with ``TableHandle``"""
        if not isinstance(value, DataHandle):
            raise TypeError(
                f"Can only add objects of type DataHandle to DataStore, not {type(value)}"
            )
        check = self.get(key)
        if check is not None and not self.allow_overwrite:
            raise ValueError(
                f"DataStore already has an item with key {key},"
                "of type {type(check)}, created by {check.creator}"
            )
        dict.__setitem__(self, key, value)
        return value

    def __getattr__(self, key):
        """Allow attribute-like parameter access"""
        try:
            return self.__getitem__(key)
        except KeyError as msg:
            # Kludge to get docstrings to work
            if key in ["__objclass__"]:  # pragma: no cover
                return None
            raise KeyError from msg

    def __setattr__(self, key, value):
        """Allow attribute-like parameter setting"""
        return self.__setitem__(key, value)

    def add_data(self, key, data, handle_class, path=None, creator="DataStore"):
        """Create a handle for some data, and insert it into the DataStore"""
        handle = handle_class(key, path=path, data=data, creator=creator)
        self[key] = handle
        return handle

    def add_handle(self, key, handle_class, path, creator="DataStore"):
        """Create a handle for some data, and insert it into the DataStore"""
        handle = handle_class(key, path=path, data=None, creator=creator)
        self[key] = handle
        return handle

    def read_file(self, key, handle_class, path, creator="DataStore", **kwargs):
        """Create a handle, use it to read a file, and insert it into the DataStore"""
        handle = handle_class(key, path=path, data=None, creator=creator)
        handle.read(**kwargs)
        self[key] = handle
        return handle

    def read(self, key, force=False, **kwargs):
        """Read the data associated to a particular key"""
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to read data {key} because {msg}") from msg
        return handle.read(force, **kwargs)

    def open(self, key, mode="r", **kwargs):
        """Open and return the file associated to a particular key"""
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to open data {key} because {msg}") from msg
        return handle.open(mode=mode, **kwargs)

    def write(self, key, **kwargs):
        """Write the data associated to a particular key"""
        try:
            handle = self[key]
        except KeyError as msg:
            raise KeyError(f"Failed to write data {key} because {msg}") from msg
        return handle.write(**kwargs)

    def write_all(self, force=False, **kwargs):
        """Write all the data in this DataStore"""
        for key, handle in self.items():
            local_kwargs = kwargs.get(key, {})
            if handle.is_written and not force:
                continue
            handle.write(**local_kwargs)


_DATA_STORE = DataStore()


def DATA_STORE():
    """Return the factory instance"""
    return _DATA_STORE



#!/usr/bin/env python
# coding: utf-8

# Prerquisites, os, and numpy
import os
import numpy as np

# Various rail modules
from rail.estimation.algos.train_z import TrainZInformer, TrainZEstimator
from rail.evaluation.single_evaluator import SingleEvaluator

from rail.core.stage import RailStage, RailPipeline

import ceci


class TrainZPipeline(RailPipeline):

    default_input_dict = dict(
        input_train='dummy.in',
        input_test='dummy.in',
    )

    def __init__(self):
        RailPipeline.__init__(self)

        DS = RailStage.data_store
        DS.__class__.allow_overwrite = True
        
        self.inform_trainz = TrainZInformer.build(
            aliases=dict(input='input_train'),
            hdf5_groupname='',
        )

        self.estimate_trainz = TrainZEstimator.build(
            aliases=dict(input='input_test'),
            connections=dict(
                model=self.inform_trainz.io.model,
            ),
            hdf5_groupname='',
        )

        self.evalute_trainz = SingleEvaluator.build(
            aliases=dict(truth='input_test'),
            connections=dict(
                input=self.estimate_trainz.io.output,
            ),
            point_estimates=['mode'],
            truth_point_estimates=["redshift"],
            metrics=["all"],
            metric_config=dict(brier=dict(limits=[0., 3.5])),
            exclude_metrics=['rmse', 'ks', 'kld', 'cvm', 'ad', 'rbpe', 'outlier'],
            hdf5_groupname='',
        )



"""
Example code that just spits out random numbers between 0 and 3
for z_mode, and Gaussian centered at z_mode with width
random_width*(1+zmode).
"""

import numpy as np
import qp

from scipy.stats import norm
from ceci.config import StageParameter as Param
from rail.estimation.estimator import CatEstimator
from rail.estimation.informer import CatInformer
from rail.core.data import TableHandle


class RandomGaussInformer(CatInformer):
    """Placeholder Informer"""

    name = "RandomGaussInformer"
    config_options = CatInformer.config_options.copy()

    def run(self):
        self.add_data("model", np.array([None]))


class RandomGaussEstimator(CatEstimator):
    """Random CatEstimator"""

    name = "RandomGaussEstimator"
    inputs = [("input", TableHandle)]
    config_options = CatEstimator.config_options.copy()
    config_options.update(
        rand_width=Param(float, 0.025, "ad hock width of PDF"),
        rand_zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
        rand_zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
        nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
        seed=Param(int, 87, msg="random seed"),
        column_name=Param(
            str,
            "mag_i_lsst",
            msg="name of a column that has the "
            "correct number of galaxies to find length of",
        ),
    )

    def __init__(self, args, **kwargs):
        """Constructor:
        Do CatEstimator specific initialization"""
        super().__init__(args, **kwargs)
        self.zgrid = None

    def _process_chunk(self, start, end, data, first):
        pdf = []
        # allow for either format for now
        numzs = len(data[self.config.column_name])
        rng = np.random.default_rng(seed=self.config.seed + start)
        zmode = np.round(rng.uniform(0.0, self.config.rand_zmax, numzs), 3)
        widths = self.config.rand_width * (1.0 + zmode)
        self.zgrid = np.linspace(
            self.config.rand_zmin, self.config.rand_zmax, self.config.nzbins
        )
        for i in range(numzs):
            pdf.append(norm.pdf(self.zgrid, zmode[i], widths[i]))
        qp_d = qp.Ensemble(
            qp.stats.norm,  # pylint: disable=no-member
            data=dict(
                loc=np.expand_dims(zmode, -1),  # pylint: disable=no-member
                scale=np.expand_dims(widths, -1),
            ),
        )
        qp_d.set_ancil(dict(zmode=zmode))
        self._do_chunk_output(qp_d, start, end, first)



"""Rail-specific data management"""

from typing import Any
import pickle


class Model:
    """Class to act as wrapper for ML models """

    def __init__(
        self,
        data: Any,
        creation_class_name: str,
        version: int=0,
        provenance: dict[str, Any] | None=None,
    ):
        """Constructor

        Parameters
        ----------
        data : Any
            Model data
        creation_class_name : str
            Name of the creation class
        version : int
            Version of the model
        provenance: dict[str, Any] | None
            Provenance infomration
        """
        self.data = data
        self.creation_class_name = creation_class_name
        self.version = version
        if provenance is not None:
            self.provenance = provenance.copy()
        else:
            self.provenance = {}

    def validate(
        self,
        creation_class_name: str | None,
        version: int | None
    ) -> None:
        """
        Parameters
        ----------
        creation_class_name : str | None
            Name of the creation class
        version : int | None
            Version of the model

        Raises
        ------
        TypeError : Either creation_class_name or version does not match
        """
        if creation_class_name is not None and creation_class_name != self.creation_class_name:
            raise TypeError(f"Model.creation_class_name does not match.  {creation_class_name} != {self.creation_class_name}")
        if version is not None and version != self.version:
            raise TypeError(f"Model.version does not match.  {version} != {self.version}")

    @classmethod
    def read(
        cls,
        path: str,
        creation_class_name: str="dummy",
        version: int=0,
        provenance: dict[str, Any] | None = None,
    ):
        """Read a model from a file.

        Note that this will promote the data to a Model if it is not already

        Parameters
        ----------
        path: str,
            File to read
        creation_class_name : str
            Name of the creation class
        version : int
            Version of the model
        provenance: dict[str, Any] | None
            Provenance infomration

        Returns
        -------
        model: Model
            Newly read Model
        """
        with open(path, 'rb') as fin:
            read_data = pickle.load(fin)

        if isinstance(read_data, Model):
            return read_data

        if provenance is None:
            provenance = {}

        return cls(read_data, creation_class_name, version, provenance)

    @classmethod
    def wrap(
        cls,
        inpath: str,
        outpath: str,
        creation_class_name: str="dummy",
        version: int=0,
        provenance: dict[str, Any] | None = None,
    ):
        """Read a model from a file and write it as a `Model` if it is not already

        Parameters
        ----------
        inpath: str,
            File to read
        outpath: str,
            File to write
        creation_class_name : str
            Name of the creation class
        version : int
            Version of the model
        provenance: dict[str, Any] | None
            Provenance information

        Returns
        -------
        model: Model
            Newly read & converted Model
        """
        the_model = cls.read(inpath, creation_class_name, version, provenance)
        the_model.write(outpath)
        return the_model

    @classmethod
    def dump(
        cls,
        obj: Any,
        path: str,
        creation_class_name: str="dummy",
        version: int=0,
        provenance: dict[str, Any] | None = None,
    ):
        """Write an object to a model file

        This will promote it to a Model if it isn't already

        Parameters
        ----------
        obj: Any
            Object to dump
        path: str,
            File to write
        creation_class_name : str
            Name of the creation class
        version : int
            Version of the model
        provenance: dict[str, Any] | None
            Provenance information

        Returns
        -------
        model: Model
            Newly converted Model
        """

        if isinstance(obj, Model):
            write_obj = obj
        else:
            if provenance is None:
                provenance = {}
            write_obj = cls(obj, creation_class_name, version, provenance)

        write_obj.write(path)
        return write_obj

    def write(
        self,
        path: str,
    ) -> None:
        """Write a model to a file

        Parameters
        ----------
        path: str,
            File to write
        """
        with open(path, "wb") as fout:
            pickle.dump(obj=self, file=fout, protocol=pickle.HIGHEST_PROTOCOL)



import os
import numpy as np
import pytest

from rail.utils.path_utils import find_rail_file
from rail.interfaces import PZFactory


def test_pz_factory():

    stage = PZFactory.build_cat_estimator_stage(
        'train_z',
        'TrainZEstimator',
        'rail.estimation.algos.train_z',
        'tests/interfaces/model_inform_trainz.pkl',
        'dummy.in',
    )

    input_file = find_rail_file('examples_data/testdata/validation_10gal.hdf5')
    
    out_single = PZFactory.estimate_single_pz(stage, {'d':np.array([1,1])})
    assert out_single.npdf == 1

    out_handle = PZFactory.run_cat_estimator_stage(
        stage,
        input_file,
    )

    check_stage = PZFactory.get_cat_estimator_stage('train_z')
    assert check_stage == stage

    with pytest.raises(KeyError):
        PZFactory.get_cat_estimator_stage('nope')    
    
    PZFactory.reset()
    assert not PZFactory._stage_dict
    
    try:
        os.unlink('inprogress_output_train_z.hdf5')
    except:
        pass
    try:
        os.unlink('output_train_z.hdf5')
    except:
        pass

    
    
    
    
        



import numpy as np
from ceci.config import StageParameter as Param
from qp.metrics.point_estimate_metric_classes import PointToPointMetric

from rail.core.data import TableHandle, QPHandle
from rail.evaluation.evaluator import Evaluator


class PointToPointEvaluator(Evaluator):
    """Evaluate the performance of a photo-z estimator against reference point estimate"""

    name = "PointToPointEvaluator"
    config_options = Evaluator.config_options.copy()
    config_options.update(
        hdf5_groupname=Param(
            str, "photometry", required=False, msg="HDF5 Groupname for truth table."
        ),
        reference_dictionary_key=Param(
            str,
            "redshift",
            required=False,
            msg="The key in the `truth` dictionary where the redshift data is stored.",
        ),
        point_estimate_key=Param(
            str, "zmode", required=False, msg="The key in the point estimate table."
        ),
    )
    inputs = [("input", QPHandle), ("truth", TableHandle)]

    metric_base_class = PointToPointMetric

    def _process_chunk(self, data_tuple, first):
        start = data_tuple[0]
        end = data_tuple[1]
        estimate_data = np.squeeze(data_tuple[2].ancil[self.config.point_estimate_key])
        reference_data = data_tuple[3][self.config.reference_dictionary_key]

        self._process_all_chunk_metrics(
            estimate_data, reference_data, start, end, first
        )

    def _process_all(self, data_tuple):
        estimate_data = np.squeeze(data_tuple[0].ancil[self.config.point_estimate_key])
        reference_data = data_tuple[1][self.config.hdf5_groupname][
            self.config.reference_dictionary_key
        ]

        self._process_all_metrics(estimate_data, reference_data)




from qp import Ensemble
from ceci.stage import PipelineStage
from rail.core.stage import RailStage
from rail.core.data import DataHandle
from rail.estimation.estimator import CatEstimator


class PZFactory:
    """ Factory class to provide a unified interface to 
    rail p(z) estimation algorithms.
    """
    
    _stage_dict = {}

    @classmethod
    def reset(cls):
        """ Reset the dictionary of cached stage objects """
        cls._stage_dict = {}

    @classmethod
    def build_stage_instance(
        cls,
        stage_name: str, 
        stage_class: type,
        model_path: str,
        data_path: str = 'none',
        **config_params: dict,
    ) -> CatEstimator:
        """ Build and configure an estimator that can evalute
        p(z) given an input catalog

        Parameters
        ----------
        stage_name: str
            Name of the stage instance, used to construct output file name

        stage_class: type
            Python class for the stage

        model_path: str
            Path to the model file used by this estimator

        data_path: str
            Path to the input data, defaults to 'none' 

        config_params: dict
            Configuration parameters for the stage

        Returns
        -------
        stage_obj: CatEstimator
            Newly constructed and configured Estimator instance
        """
        stage_obj = stage_class.make_stage(name=stage_name, model=model_path, input=data_path, **config_params)
        cls._stage_dict[stage_name] = stage_obj
        return stage_obj
        
    @classmethod
    def build_cat_estimator_stage(
        cls,
        stage_name: str,
        class_name: str,
        module_name: str,
        model_path: str,
        data_path: str = 'none',
        **config_params: dict,
    ) -> CatEstimator:
        """ Build and configure an estimator that can evalute
        p(z) given an input catalog

        Parameters
        ----------
        stage_name: str
            Name of the stage instance, used to construct output file name

        class_name: str
            Name of the class, e.g., TrainZEstimator, used to find the class

        module_name: str
            Name of the python module that constains the class, used to import it

        model_path: str
            Path to the model file used by this estimator

        data_path: str
            Path to the input data, defaults to 'none' 

        config_params: dict
            Configuration parameters for the stage

        Returns
        -------
        stage_obj: CatEstimator
            Newly constructed and configured Estimator instance
        """
        stage_class = PipelineStage.get_stage(class_name, module_name)
        return cls.build_stage_instance(stage_name, stage_class, model_path, data_path, **config_params)

    @classmethod
    def get_cat_estimator_stage(
        cls,
        stage_name: str,
    ) -> CatEstimator:
        """ Return a cached p(z) estimator """
        try:
            return cls._stage_dict[stage_name]
        except KeyError as msg:
            raise KeyError(
                f"Could not find stage named {stage_name}, did you build it?"
                f"Existing stages are: {list(cls._stage_dict.keys())}"
            ) from msg

    @staticmethod
    def run_cat_estimator_stage(
        stage_obj: CatEstimator,
        data_path: str,
    ) -> DataHandle:
        """ Run a p(z) estimator on an input data file

        Parameters
        ----------
        stage_obj: CatEstimator
            Object that will do the estimation

        Returns
        -------
        data_handle: DataHandle
            Object that can give access to the data
        """        
        RailStage.data_store.clear()        
        handle = stage_obj.get_handle('input', path=data_path, allow_missing=True)        
        return stage_obj.estimate(handle)

    @staticmethod
    def estimate_single_pz(
        stage_obj: CatEstimator,
        data_table: dict,
        input_size: int=1,
    ) -> Ensemble:
        """ Run a p(z) estimator on some objects

        Parameters
        ----------
        stage_obj: CatEstimator
            Object that will do the estimation

        data_table: dict
            Input data presented as dict of numpy arrays objects

        input_size: int
            Number of objects in the input table

        Returns
        -------
        pz : qp.Ensemble
            Output pz
        """        
        RailStage.data_store.clear()
        if stage_obj.model is None:
            stage_obj.open_model(**stage_obj.config)
        stage_obj._input_length = input_size
        stage_obj._process_chunk(0, input_size, data_table, True)
        return stage_obj._output_handle.data
        
   



import numpy as np


from rail.core import common_params


class CatalogConfigBase:

    sub_classes = {}

    tag = None
    bandlist = ''
    maglims = []
    band_template = ''
    band_err_template = ''
    ref_band = ''
    redshift_col = ''

    _active_tag = None
    _active_class = None
    
    def __init_subclass__(cls, **kwargs):        
        cls.sub_classes[cls.tag] = cls

    @classmethod
    def active_tag(cls):
        return cls._active_tag

    @classmethod
    def active_class(cls):
        return cls._active_class

    @classmethod
    def apply(cls, tag):
        cls._active_tag = tag
        cls._active_class = cls.sub_classes[tag]
        cls._active_class._apply()
        
    @classmethod
    def _apply(cls):
        base_dict = cls._build_base_dict()
        common_params.set_param_defaults(**base_dict)
        cls._apply_hook()
        
    @classmethod
    def _apply_hook(cls):
        return 

    @classmethod
    def band_name_dict(cls):
        return {band: cls.band_template.format(band=band) for band in cls.bandlist}

    @classmethod
    def _build_maglim_dict(cls):
        maglim_dict = {}
        for band, limx in zip(cls.bandlist, cls.maglims):
            maglim_dict[cls.band_template.format(band=band)] = limx
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls):
        a_env_dict = {}
        for band, limx in zip(cls.bandlist, cls.a_env):
            a_env_dict[cls.band_template.format(band=band)] = limx
        return a_env_dict

    @classmethod
    def _build_band_names(cls):
        return [cls.band_template.format(band=band) for band in cls.bandlist]

    @classmethod
    def _build_band_err_names(cls):
        return [cls.band_err_template.format(band=band) for band in cls.bandlist]

    @classmethod
    def _build_ref_band(cls, ref_band='i'):
        return cls.band_template.format(band=ref_band)
    
    @classmethod
    def _build_base_dict(cls):
        base_dict = {}
        base_dict.update(
            bands=cls._build_band_names(),
            err_bands=cls._build_band_err_names(),
            mag_limits=cls._build_maglim_dict(),
            band_a_env=cls._build_a_env_dict(),
            ref_band=cls._build_ref_band(cls.ref_band),
            redshift_col=cls.redshift_col,
            nondetect_val=np.nan,
            hdf5_groupname=cls.hdf5_groupname,
        )
        return base_dict

        
class HscCatalogConfig(CatalogConfigBase):

    tag = 'hsc'
    bandlist = 'grizy'
    maglims = [27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [3.64, 2.70, 2.06, 1.58, 1.31]    
    band_template = 'HSC{band}_cmodel_dered'
    band_err_template = '{band}_cmodel_magerr' 
    ref_band = 'i'
    redshift_col = 'specz_redshift'
    hdf5_groupname = ''

    
class Dc2CatalogConfig(CatalogConfigBase):

    tag = 'dc2'
    bandlist = 'ugrizy'
    maglims = [24., 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]    
    band_template = "mag_{band}_lsst"
    band_err_template = "mag_err_{band}_lsst"
    ref_band = 'i'
    redshift_col='true_redshift'
    hdf5_groupname = ''


class RubinCatalogConfig(CatalogConfigBase):

    tag = 'rubin'
    bandlist = 'ugrizy'
    maglims = [24., 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template ="LSST_obs_{band}_err"
    ref_band = 'i'
    redshift_col='true_redshift'
    hdf5_groupname = ''


class RomanRubinCatalogConfig(CatalogConfigBase):

    tag = 'roman_rubin'
    bandlist = 'ugrizy'
    maglims = [24., 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template ="LSST_obs_{band}_err"
    ref_band = 'i'
    redshift_col='redshift'
    hdf5_groupname = ''



class RomanPlusRubinCatalogConfig(CatalogConfigBase):

    tag = 'roman_plus_rubin'
    bandlist = 'ugrizy'
    maglims = [24., 27.66, 27.25, 26.6, 26.24, 25.35]
    a_env = [4.81, 3.64, 2.70, 2.06, 1.58, 1.31]
    band_template = "LSST_obs_{band}"
    band_err_template ="LSST_obs_{band}_err"
    ref_band = 'i'
    redshift_col='redshift'
    hdf5_groupname = ''

    @classmethod
    def band_name_dict(cls):
        bands = super().band_name_dict()
        bands['f'] = 'ROMAN_obs_F184'
        bands['h'] = 'ROMAN_obs_H158'
        bands['j'] = 'ROMAN_obs_J129'
        bands['y'] = 'ROMAN_obs_Y106'
        return bands
        
    @classmethod
    def _build_maglim_dict(cls):
        maglim_dict = super()._build_maglim_dict()
        maglim_dict['ROMAN_obs_F184'] = 27.5
        maglim_dict['ROMAN_obs_H158'] = 28.1
        maglim_dict['ROMAN_obs_J129'] = 27.8
        maglim_dict['ROMAN_obs_Y106'] = 27.6
        return maglim_dict

    @classmethod
    def _build_a_env_dict(cls):
        a_env_dict = super()._build_a_env_dict()
        a_env_dict['ROMAN_obs_F184'] = 1.1
        a_env_dict['ROMAN_obs_H158'] = 1.2
        a_env_dict['ROMAN_obs_J129'] = 1.3
        a_env_dict['ROMAN_obs_Y106'] = 1.4
        return a_env_dict

    @classmethod
    def _build_band_names(cls):
        bands = [cls.band_template.format(band=band) for band in cls.bandlist]
        bands += ['ROMAN_obs_F184', 'ROMAN_obs_H158', 'ROMAN_obs_J129', 'ROMAN_obs_Y016']
        return bands

    @classmethod
    def _build_band_err_names(cls):
        band_errs = [cls.band_err_template.format(band=band) for band in cls.bandlist]
        band_errs += ['ROMAN_obs_F184_err', 'ROMAN_obs_H158_err', 'ROMAN_obs_J129_err', 'ROMAN_obs_Y106_err']
        return band_errs

    @classmethod
    def _build_ref_band(cls, ref_band='i'):
        return cls.band_template.format(band=ref_band)

    
apply_defaults = CatalogConfigBase.apply



import os
import numpy as np
import pytest
import qp

from rail.utils.path_utils import RAILDIR
from rail.core.stage import RailStage
from rail.core.data import TableHandle
from rail.estimation.algos.true_nz import TrueNZHistogrammer


DS = RailStage.data_store
DS.__class__.allow_overwrite = True

true_nz_file = "src/rail/examples_data/testdata/validation_10gal.hdf5"
tomo_file = "src/rail/examples_data/testdata/output_tomo.hdf5"


def test_true_nz():
    DS.clear()
    true_nz = DS.read_file('true_nz', path=true_nz_file, handle_class=TableHandle)
    tomo_bins = DS.read_file('tomo_bins', path=tomo_file, handle_class=TableHandle)
    
    nz_hist = TrueNZHistogrammer.make_stage(
        name='true_nz',
        hdf5_groupname='photometry',
        redshift_col='redshift',
        zmin=0.0,
        zmax=3.0,
        nzbins=301,            
    )
    out_hist = nz_hist.histogram(true_nz, tomo_bins)

    check_ens = qp.read(out_hist.path)
    assert check_ens.ancil['n_total'][0] == 10

    check_vals = [0, 5, 0, 0, 0]
    
    for i in range(5):
        nz_hist = TrueNZHistogrammer.make_stage(
            name='true_nz',
            hdf5_groupname='photometry',
            redshift_col='redshift',
            zmin=0.0,
            zmax=3.0,
            nzbins=301,            
            selected_bin=i,
        )
        out_hist = nz_hist.histogram(true_nz, tomo_bins)
        check_ens = qp.read(out_hist.path)
        assert check_ens.ancil['n_total'][0] == check_vals[i]

        



"""
Abstract base classes defining classifiers.
"""

import gc
from rail.core.data import QPHandle, TableHandle, ModelHandle, Hdf5Handle
from rail.core.common_params import SHARED_PARAMS
from rail.core.stage import RailStage


class CatClassifier(RailStage):  # pragma: no cover
    """The base class for assigning classes to catalogue-like table.

    Classifier uses a generic "model", the details of which depends on the sub-class.

    CatClassifier take as "input" a catalogue-like table, assign each object into
    a tomographic bin, and provide as "output" a tabular data which can be appended
    to the catalogue.
    """

    name = "CatClassifier"
    config_options = RailStage.config_options.copy()
    config_options.update(
        chunk_size=SHARED_PARAMS,
        hdf5_groupname=SHARED_PARAMS,
    )
    inputs = [("model", ModelHandle), ("input", TableHandle)]
    outputs = [("output", TableHandle)]

    def __init__(self, args, **kwargs):
        """Initialize Classifier"""
        super().__init__(args, **kwargs)
        self._output_handle = None
        self.model = None
        if not isinstance(args, dict):  # pragma: no cover
            args = vars(args)
        self.open_model(**args)

    def open_model(self, **kwargs):
        """Load the model and/or attach it to this Classifier

        Parameters
        ----------
        model : `object`, `str` or `ModelHandle`
            Either an object with a trained model,
            a path pointing to a file that can be read to obtain the trained model,
            or a `ModelHandle` providing access to the trained model.

        Returns
        -------
        self.model : `object`
            The object encapsulating the trained model.
        """
        model = kwargs.get("model", None)
        if model is None or model == "None":
            self.model = None
            return self.model
        if isinstance(model, str):
            self.model = self.set_data("model", data=None, path=model)
            self.config["model"] = model
            return self.model
        if isinstance(model, ModelHandle):
            if model.has_path:
                self.config["model"] = model.path
        self.model = self.set_data("model", model)
        return self.model

    def classify(self, input_data):
        """The main run method for the classifier, should be implemented
        in the specific subclass.

        This will attach the input_data to this `CatClassifier`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this Classifier
        by using `self.add_data('output', output_data)`.

        Finally, this will return a TableHandle providing access to that output data.

        Parameters
        ----------
        input_data : `dict`
            A dictionary of all input data

        Returns
        -------
        output: `dict`
            Class assignment for each galaxy.
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        return self.get_handle("output")


class PZClassifier(RailStage):
    """The base class for assigning classes (tomographic bins) to per-galaxy PZ
    estimates.

    PZClassifier takes as "input" a `qp.Ensemble` with per-galaxy PDFs, and
    provides as "output" tabular data which can be appended to the catalogue.
    """

    name = "PZClassifier"
    config_options = RailStage.config_options.copy()
    config_options.update(chunk_size=SHARED_PARAMS)
    inputs = [("input", QPHandle)]
    outputs = [("output", Hdf5Handle)]

    def __init__(self, args, **kwargs):
        """Initialize the PZClassifier.

        Parameters
        ----------
        args : dict
            Configuration arguments for the classifier.
        comm : MPI.Comm, optional
            MPI communicator for parallel processing.
        """
        super().__init__(args, **kwargs)
        self._output_handle = None

    def classify(self, input_data):
        """The main run method for the classifier, should be implemented
        in the specific subclass.

        This will attach the input_data to this `PZClassifier`
        (for introspection and provenance tracking).

        Then it will call the run() and finalize() methods, which need to
        be implemented by the sub-classes.

        The run() method will need to register the data that it creates to this
        Classifier by using `self.add_data('output', output_data)`.

        The run() method relies on the _process_chunk() method, which should be
        implemented by subclasses to perform the actual classification on each
        chunk of data. The results from each chunk are then combined in the
        _finalize_run() method. (Alternatively, override run() in a subclass to
        perform the classification without parallelization.)

        Finally, this will return a TableHandle providing access to that output data.

        Parameters
        ----------
        input_data : `qp.Ensemble`
            Per-galaxy p(z), and any ancilary data associated with it

        Returns
        -------
        output: `dict`
            Class assignment for each galaxy, typically in the form of a
            dictionary with IDs and class labels.
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        return self.get_handle("output")

    def _finalize_run(self):
        """Finalize the classification process after processing all chunks."""
        self._output_handle.finalize_write()

    def _process_chunk(self, start, end, data, first):
        """Process a chunk of data.

        This method should be implemented in subclasses to perform the actual
        classification on a chunk of data.

        Parameters
        ----------
        start : int
            The starting index of the chunk.
        end : int
            The ending index of the chunk.
        data : qp.Ensemble
            The data chunk to be processed.
        first : bool
            True if this is the first chunk, False otherwise.
        """
        raise NotImplementedError(
            f"{self.name}._process_chunk is not implemented"
        )  # pragma: no cover

    def _do_chunk_output(self, class_id, start, end, first):
        """Handle the output of a processed chunk.

        Parameters
        ----------
        class_id : dict
            The classification results for the chunk.
        start : int
            The starting index of the chunk.
        end : int
            The ending index of the chunk.
        first : bool
            True if this is the first chunk, False otherwise.
        """
        if first:
            self._output_handle = self.add_handle("output", data=class_id)
            self._output_handle.initialize_write(
                self._input_length, communicator=self.comm
            )
        self._output_handle.set_data(class_id, partial=True)
        self._output_handle.write_chunk(start, end)

    def run(self):
        """Processes the input data in chunks and performs classification.

        This method iterates over chunks of the input data, calling the
        _process_chunk method for each chunk to perform the actual classification.

        The _process_chunk method should be implemented by subclasses to define
        the specific classification logic.
        """
        test_data = self.get_data("input")

        iterator = self.input_iterator("input")
        first = True

        for start, end, test_data in iterator:
            # print(f"Process {self.rank} running estimator on chunk {start} - {end}")
            self._process_chunk(start, end, test_data, first)
            first = False
            # Running garbage collection manually seems to be needed
            # to avoid memory growth for some estimators
            gc.collect()
        self._finalize_run()



import pickle
import numpy as np

import pytest

from rail.core.model import Model
from rail.estimation.algos.train_z import trainZmodel


def pickle_dump(obj, path):
    with open(path, 'wb') as fout:
        pickle.dump(obj=obj, file=fout, protocol=pickle.HIGHEST_PROTOCOL)


def test_model():

    array_data = np.array([None])
    dict_data = dict(a=5, b=6)
    train_z_data = trainZmodel(np.linspace(0., 3., 301), np.ones(301), 1.5)
    model = Model(array_data, 'dummy', 0, dict(alice='bob'))
    model_2 = Model(array_data, 'dummy', 0)

    pickle_dump(array_data, 'array_raw.pickle')
    read_array_data = Model.read('array_raw.pickle', 'dummy', 0, dict(alice='bob'))
    read_array_data_2 = Model.read('array_raw.pickle', 'dummy', 0)

    Model.wrap('array_raw.pickle', 'array_wrap.pickle', 'dummy', 0, dict(alice='bob'))
    
    Model.dump(array_data, 'array_data.pickle', 'dummy', 0)
    Model.dump(dict_data, 'dict_data.pickle', 'dummy', 0, dict(alice='bob'))
    Model.dump(train_z_data, 'train_z_data.pickle', 'dummy', 0, dict(alice='bob'))
    Model.dump(model, 'model_data.pickle')

    check_array_data = Model.read('array_data.pickle')
    check_dict_data = Model.read('dict_data.pickle')
    check_train_z_data = Model.read('train_z_data.pickle')
    check_model_data = Model.read('model_data.pickle')

    Model.validate(check_array_data, 'dummy', 0)
    with pytest.raises(TypeError):
        Model.validate(check_array_data, 'dummies', 0)
  
    with pytest.raises(TypeError):
        Model.validate(check_array_data, 'dummy', 1)
  

    

    



import rail

from rail.core import RailEnv
from rail.core.stage import RailStage

from rail.estimation.estimator import CatEstimator
from rail.estimation.classifier import CatClassifier, PZClassifier
from rail.estimation.summarizer import CatSummarizer, PZSummarizer, SZPZSummarizer

from rail.estimation.algos.naive_stack import NaiveStackInformer, NaiveStackSummarizer
from rail.estimation.algos.random_gauss import RandomGaussInformer, RandomGaussEstimator
from rail.estimation.algos.point_est_hist import (
    PointEstHistInformer,
    PointEstHistSummarizer,
)
from rail.estimation.algos.train_z import TrainZInformer, TrainZEstimator
from rail.estimation.algos.var_inf import VarInfStackInformer, VarInfStackSummarizer
from rail.estimation.algos.uniform_binning import UniformBinningClassifier
from rail.estimation.algos.equal_count import EqualCountClassifier
from rail.estimation.algos.true_nz import TrueNZHistogrammer

from rail.creation.degrader import Degrader

from rail.creation.degraders.addRandom import AddColumnOfRandom
from rail.creation.degraders.quantityCut import QuantityCut

from rail.creation.engine import Modeler, Creator, PosteriorCalculator
from rail.creation.noisifier import Noisifier
from rail.creation.selector import Selector

from rail.evaluation.dist_to_dist_evaluator import DistToDistEvaluator
from rail.evaluation.dist_to_point_evaluator import DistToPointEvaluator
from rail.evaluation.evaluator import Evaluator, OldEvaluator
from rail.evaluation.point_to_point_evaluator import PointToPointEvaluator
from rail.evaluation.single_evaluator import SingleEvaluator

from rail.tools.table_tools import ColumnMapper, RowSelector, TableConverter

__all__ = [
    "CatEstimator",
    "CatClassifier",
    "PZClassifier",
    "CatSummarizer",
    "PZSummarizer",
    "SZPZSummarizer",
    "NaiveStackInformer",
    "NaiveStackSummarizer",
    "RandomGaussInformer",
    "RandomGaussEstimator",
    "PointEstHistInformer",
    "PointEstHistSummarizer",
    "TrainZInformer",
    "TrainZEstimator",
    "VarInfStackInformer",
    "VarInfStackSummarizer",
    "UniformBinningClassifier",
    "EqualCountClassifier",
    "TrueNZHistogrammer",
    "Degrader",
    "AddColumnOfRandom",
    "QuantityCut",
    "Modeler",
    "Creator",
    "PosteriorCalculator",
    "Noisifier",
    "Selector",
    "DistToDistEvaluator",
    "DistToPointEvaluator",
    "Evaluator",
    "OldEvaluator",
    "PointToPointEvaluator",
    "SingleEvaluator",
    "ColumnMapper",
    "RowSelector",
    "TableConverter",
]


def import_and_attach_all():
    """Import all the packages in the rail ecosystem and attach them to this module"""
    RailEnv.import_all_packages()
    RailEnv.attach_stages(rail.stages)
    for xx in RailStage.pipeline_stages:
        rail.stages.__all__.append(xx)
    



import os
from rail.utils.testing_utils import build_and_read_pipeline

import pytest

@pytest.mark.parametrize(
    "pipeline_class",
    [
        'rail.pipelines.estimation.train_z_pipeline.TrainZPipeline',
    ]
)
def test_build_and_read_pipeline(pipeline_class):
    build_and_read_pipeline(pipeline_class)




# This is a deprecative path for importing RAILDIR, it is only for the transition period
# to move RAILDIR and find_rail_file to utils.path_utils

from rail.utils.path_utils import find_rail_file
from rail.utils.path_utils import RAILDIR

__all__ = [
    'find_rail_file',
    'RAILDIR',
]



"""
Abstract base classes defining Estimators of individual galaxy redshift uncertainties.
"""

import gc

from ceci.config import StageParameter as Param
from rail.core.common_params import SHARED_PARAMS
from rail.core.data import TableHandle, QPHandle, ModelHandle
from rail.core.stage import RailStage

from rail.core.point_estimation import PointEstimationMixin

# for backwards compatibility, to avoid break stuff that imports it from here
from rail.estimation.informer import CatInformer  # pylint: disable=unused-import

class CatEstimator(RailStage, PointEstimationMixin):
    """The base class for making photo-z posterior estimates from catalog-like inputs
    (i.e., tables with fluxes in photometric bands among the set of columns)

    Estimators use a generic "model", the details of which depends on the sub-class.

    Estimators take as "input" tabular data, apply the photo-z estimation and
    provide as "output" a ``QPEnsemble``, with per-object p(z).

    """

    name = "CatEstimator"
    config_options = RailStage.config_options.copy()
    config_options.update(
        chunk_size=SHARED_PARAMS,
        hdf5_groupname=SHARED_PARAMS,
        zmin=SHARED_PARAMS,
        zmax=SHARED_PARAMS,
        nzbins=SHARED_PARAMS,
    )
    config_options.update(
        **PointEstimationMixin.config_options.copy(),
    )
    inputs = [("model", ModelHandle), ("input", TableHandle)]
    outputs = [("output", QPHandle)]

    def __init__(self, args, **kwargs):
        """Initialize Estimator"""
        super().__init__(args, **kwargs)
        self._output_handle = None
        self.model = None

    def open_model(self, **kwargs):
        """Load the model and attach it to this Estimator

        Parameters
        ----------
        model : ``object``, ``str`` or ``ModelHandle``
            Either an object with a trained model, a path pointing to a file
            that can be read to obtain the trained model, or a `ModelHandle`
            providing access to the trained model.

        Returns
        -------
        self.model : ``object``
            The object encapsulating the trained model.
        """
        model = kwargs.get("model", None)
        if model is None or model == "None":
            self.model = None
            return self.model
        if isinstance(model, str):
            self.model = self.set_data("model", data=None, path=model)
            self.config["model"] = model
            return self.model
        if isinstance(model, ModelHandle):
            if model.has_path:
                self.config["model"] = model.path
        self.model = self.set_data("model", model)
        return self.model

    def estimate(self, input_data):
        """The main interface method for the photo-z estimation

        This will attach the input data (defined in ``inputs`` as "input") to this
        ``Estimator`` (for introspection and provenance tracking). Then call the
        ``run()`` and ``finalize()`` methods.

        The run method will call ``_process_chunk()``, which needs to be implemented
        in the subclass, to process input data in batches. See ``RandomGaussEstimator``
        for a simple example.

        Finally, this will return a ``QPHandle`` for access to that output data.

        Parameters
        ----------
        input_data : ``dict`` or ``ModelHandle``
            Either a dictionary of all input data or a ``ModelHandle`` providing
            access to the same

        Returns
        -------
        output: ``QPHandle``
            Handle providing access to QP ensemble with output data
        """
        self.set_data("input", input_data)
        self.run()
        self.finalize()
        results = self.get_handle("output")
        results.read(force = True)
        return results

    def run(self):
        self.open_model(**self.config)

        iterator = self.input_iterator("input")
        first = True
        self._initialize_run()
        self._output_handle = None
        for s, e, test_data in iterator:
            print(f"Process {self.rank} running estimator on chunk {s} - {e}")
            self._process_chunk(s, e, test_data, first)
            first = False
            # Running garbage collection manually seems to be needed
            # to avoid memory growth for some estimators
            gc.collect()
        self._finalize_run()

    def _initialize_run(self):
        self._output_handle = None

    def _finalize_run(self):
        self._output_handle.finalize_write()

    def _process_chunk(self, start, end, data, first):
        raise NotImplementedError(
            f"{self.name}._process_chunk is not implemented"
        )  # pragma: no cover

    def _do_chunk_output(self, qp_dstn, start, end, first):
        qp_dstn = self.calculate_point_estimates(qp_dstn)
        if first:
            self._output_handle = self.add_handle("output", data=qp_dstn)
            if self.config.output_mode != 'return':
                self._output_handle.initialize_write(
                    self._input_length, communicator=self.comm
                )
        self._output_handle.set_data(qp_dstn, partial=True)
        if self.config.output_mode != 'return':
            self._output_handle.write_chunk(start, end)
        return qp_dstn



import os
import click
import pprint
import yaml

from rail.core import __version__
from rail.cli.rail import options, scripts
from rail.interfaces.pz_factory import PZFactory
from rail.interfaces.tool_factory import ToolFactory
from rail.utils import catalog_utils
import ceci


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """RAIL utility scripts"""


@cli.command()
@options.outdir(default="docs")
@options.clear_output()
@options.dry_run()
@options.inputs()
@options.skip()
def render_nb(outdir, clear_output, dry_run, inputs, skip, **_kwargs):
    """Render jupyter notebooks"""
    return scripts.render_nb(outdir, clear_output, dry_run, inputs, skip)


@cli.command()
@options.outdir(default="..")
@options.git_mode()
@options.dry_run()
@options.package_file()
def clone_source(outdir, git_mode, dry_run, package_file, **_kwargs):
    """Install packages from source"""
    scripts.clone_source(outdir, git_mode, dry_run, package_file)
    return 0


@cli.command()
@options.outdir(default="..")
@options.dry_run()
@options.package_file()
def update_source(outdir, dry_run, package_file, **_kwargs):
    """Update packages from source"""
    scripts.update_source(outdir, dry_run, package_file)
    return 0


@cli.command()
@options.outdir(default="..")
@options.dry_run()
@options.from_source()
@options.package_file()
def install(outdir, dry_run, from_source, package_file, **_kwargs):
    """Install rail packages one by one, to be fault tolerant"""
    scripts.install(outdir, from_source, dry_run, package_file)
    return 0


@cli.command()
@options.outdir(default="..")
@options.print_all()
@options.print_packages()
@options.print_namespaces()
@options.print_modules()
@options.print_tree()
@options.print_stages()
def info(**kwargs):
    """Print information about the rail ecosystem"""
    scripts.info(**kwargs)
    return 0


@cli.command()
@options.bpz_demo_data()
@options.verbose_download()
def get_data(verbose, **kwargs):
    """Downloads data from NERSC (if not already found)"""
    scripts.get_data(verbose, **kwargs)
    return 0


@cli.command()
@options.stage_name()
@options.stage_class()
@options.stage_module()
@options.stages_config()
@options.model_file()
@options.catalog_tag()
@options.dry_run()
@options.input_file()
@options.params()
def estimate(stage_name, stage_class, stage_module, stages_config, model_file, catalog_tag, dry_run, input_file, params):
    """Run a pz estimation stage"""
    if catalog_tag:
        catalog_utils.apply_defaults(catalog_tag)

    if stages_config not in [None, 'none', 'None']:
        with open(stages_config) as fin:
            config_data = yaml.safe_load(fin)
            if stage_name in config_data:
                kwargs = config_data[stage_name]
            elif isinstance(config_data, dict):
                kwargs = config_data
            else:
                raise ValueError(f"Config file {stages_config} is not properly constructed")                
    else:
        kwargs = {}
        
    kwargs.update(**options.args_to_dict(params))
    
    stage = PZFactory.build_cat_estimator_stage(
        stage_name=stage_name,
        class_name=stage_class,
        module_name=stage_module,
        model_path=model_file,
        data_path='dummy.in',
        **kwargs,
    )

    if dry_run:
        print(f"Running stage {stage.name} of type {type(stage)} on {input_file}")
        print("Stage config:   ")
        print_dict = stage.config.to_dict().copy()
        pprint.pprint(print_dict)
        return 0
        
    output = PZFactory.run_cat_estimator_stage(
        stage,
        data_path=input_file,
    )


@cli.command()
@options.pipeline_class()
@options.output_yaml()
@options.catalog_tag()
@options.stages_config()
@options.outdir()
@options.inputs()
def build_pipe(pipeline_class, output_yaml, catalog_tag, stages_config, outdir, inputs):
    """Build a pipeline yaml file"""
    input_dict = {}
    for input_ in inputs:
        tokens = input_.split('=')
        assert len(tokens) == 2
        input_dict[tokens[0]] = tokens[1]
    scripts.build_pipeline(pipeline_class, output_yaml, catalog_tag, input_dict, stages_config, outdir)
    return 0


@cli.command()
@options.pipeline_yaml()
@options.stage_name()
@options.dry_run()
@options.inputs()
def run_stage(pipeline_yaml, stage_name, dry_run, inputs):
    """Run a pipeline stage"""
    pipe = ceci.Pipeline.read(pipeline_yaml)
    input_dict = {}
    for input_ in inputs:
        tokens = input_.split('=')
        assert len(tokens) == 2
        input_dict[tokens[0]] = tokens[1]
    com = pipe.generate_stage_command(stage_name, **input_dict)
    if dry_run:
        print(com)
    else:
        os.system(com)
    return 0


@cli.command()
@options.stage_name()
@options.stage_class()
@options.stage_module()
@options.dry_run()
@options.input_file()
def run_tool(stage_name, stage_class, stage_module, dry_run, input_file):
    """Run a pz estimation stage"""
    stage = ToolFactory.build_tool_stage(
        stage_name=stage_name,
        class_name=stage_class,
        module_name=stage_module,
        data_path='dummy.in',
    )

    output = ToolFactory.run_tool_stage(
        stage,
        data_path=input_file,
    )



""" Parameters that are shared between stages """

from ceci.config import StageParameter as Param
from ceci.config import StageConfig


lsst_bands = "ugrizy"
lsst_mag_cols = [f"mag_{band}_lsst" for band in lsst_bands]
lsst_mag_err_cols = [f"mag_err_{band}_lsst" for band in lsst_bands]
lsst_def_maglims = dict(
    mag_u_lsst=27.79,
    mag_g_lsst=29.04,
    mag_r_lsst=29.06,
    mag_i_lsst=28.62,
    mag_z_lsst=27.98,
    mag_y_lsst=27.05,
)
# default reddening parameters for LSST
lsst_def_a_env = dict(
    mag_u_lsst=4.81,
    mag_g_lsst=3.64,
    mag_r_lsst=2.70,
    mag_i_lsst=2.06,
    mag_z_lsst=1.58,
    mag_y_lsst=1.31,
)


SHARED_PARAMS = StageConfig(
    hdf5_groupname=Param(
        str, "photometry", msg="name of hdf5 group for data, if None, then set to ''"
    ),
    chunk_size=Param(int, 10000, msg="Number of object per chunk for parallel processing"),
    zmin=Param(float, 0.0, msg="The minimum redshift of the z grid"),
    zmax=Param(float, 3.0, msg="The maximum redshift of the z grid"),
    nzbins=Param(int, 301, msg="The number of gridpoints in the z grid"),
    dz=Param(float, 0.01, msg="delta z in grid"),
    nondetect_val=Param(
        float, 99.0, msg="value to be replaced with magnitude limit for non detects"
    ),
    bands=Param(
        list, lsst_mag_cols, msg="Names of columns for magnitgude by filter band"
    ),
    err_bands=Param(
        list,
        lsst_mag_err_cols,
        msg="Names of columns for magnitgude errors by filter band",
    ),
    mag_limits=Param(dict, lsst_def_maglims, msg="Limiting magnitdues by filter"),
    band_a_env=Param(dict, lsst_def_a_env, msg="Redenning parameters"),
    ref_band=Param(str, "mag_i_lsst", msg="band to use in addition to colors"),
    redshift_col=Param(str, "redshift", msg="name of redshift column"),
    calculated_point_estimates=Param(
        dtype=list,
        default=[],
        msg="List of strings defining which point estimates to automatically calculate using `qp.Ensemble`."
        "Options include, 'mean', 'mode', 'median'.",
    ),
    recompute_point_estimates=Param(
        dtype=bool,
        default=False,
        msg="Force recomputation of point estimates",
    ),
)


def copy_param(param_name):
    """Return a copy of one of the shared parameters"""
    return SHARED_PARAMS.get(param_name).copy()


def set_param_default(param_name, default_value):
    """Change the default value of one of the shared parameters"""
    try:
        SHARED_PARAMS.get(param_name).set_default(default_value)
    except AttributeError as msg:  # pragma: no cover
        raise KeyError(f"No shared parameter {param_name} in SHARED_PARAMS") from msg


def set_param_defaults(**kwargs):  # pragma: no cover
    """Change the default value of several of the shared parameters"""
    for key, val in kwargs.items():
        set_param_default(key, val)



import glob
import os
import pytest
from rail.cli.rail import scripts
from rail.cli.rail.options import args_to_dict, GitMode



def test_args_to_dict():
    args = ['x=3', 'y=2']
    as_dict = args_to_dict(args)
    assert as_dict['x'] == '3'
    assert as_dict['y'] == '2'

    with pytest.raises(ValueError):
        args_to_dict(['x'])
    with pytest.raises(ValueError):
        args_to_dict(['x=3=3'])


def test_render_nb():
    nb_dir = "./tests/cli/"
    nb_files = glob.glob(os.path.join(nb_dir, "*.ipynb"))
    scripts.render_nb("docs", False, True, nb_files, skip=[])
    scripts.render_nb(
        "docs", True, True, nb_files, skip=["./tests/cli/single_number.ipynb"]
    )
    scripts.render_nb(
        "docs", True, False, nb_files, skip=["./tests/cli/single_number.ipynb"]
    )


def test_clone_source():
    scripts.clone_source("..", GitMode.ssh, True, "rail_packages.yml")
    scripts.clone_source("..", GitMode.https, True, "rail_packages.yml")
    scripts.clone_source("..", GitMode.cli, True, "rail_packages.yml")


def test_update_source():
    scripts.update_source("..", True, "rail_packages.yml")


def test_install():
    scripts.install("..", False, True, "rail_packages.yml")
    scripts.install("..", True, True, "rail_packages.yml")


def test_info():
    scripts.info(print_all=True)


def test_build_pipeline():
    scripts.build_pipeline(
        'rail.pipelines.estimation.train_z_pipeline.TrainZPipeline',
        'trainz_pipe.yaml',
        input_dict = dict(input='dummy.ibn'),
    )
